require 'rack'

require "rate_attack/version"
require 'rate_attack/cache'
require 'rate_attack/configuration'
require 'rate_attack/request'
require 'rate_attack/railtie' if defined?(::Rails)

class RateAttack
  class Error < StandardError; end

  class << self
    attr_accessor :enabled
    attr_reader :configuration

    def cache
      @cache ||= Cache.new
    end
  end

  # Set defaults
  @enabled = true
  @configuration = Configuration.new
  
  attr_reader :configuration
  
  def initialize(app)
    @app = app
    @configuration = self.class.configuration
  end

  def call(env)
    return @app.call(env) if !self.class.enabled || env["rate_attack.called"]

    env["rate_attack.called"] = true
 
    # request = RateAttack::Request.new(env)
    s_path = sanitized_path(env['PATH_INFO'])
    if s_path =~ /rateattack/
      handle_rateattack_ratelimit_info(env)
    else
      r_path = Regexp.new(s_path)
      route = wrapped_routes.filter { sanitized_path(_1.path) =~ r_path }&.first
      route ? handle_rateattack_request(env, route) : @app.call(env)
    end
  end

  def rateattack_request_key(env, info)
    request = ActionDispatch::Request.new env
    keys = []
    if info[:uniqueness].is_a?(Array)
      keys =
        info[:uniqueness].map do |item|
          item.to_s == 'ip' ? guess_ip(request) : env[item.to_s]
        end
    end
    keys.insert(0, 'RateAttack')
    keys.push(info[:action] || sanitized_path(env['PATH_INFO']))
    keys.join('::')
  end

  def get_route_ratelimit_info(env, rateattack_config)
    count =
      Rails
        .cache
        .read(rateattack_request_key(env, rateattack_config), raw: true)
        .to_i
    {
      limit: rateattack_config[:limit],
      count: count,
      remaining: (rateattack_config[:limit] - count),
      action: (rateattack_config[:action] || sanitized_path(env['PATH_INFO'])),
    }
  end

  def sanitized_path(path)
    path.gsub(%r{^\/}, '').gsub(/\(.:format\)/, '')
  end

  def handle_rateattack_ratelimit_info(env)
    ra_ratelimit_infos =
      wrapped_routes.map do
        get_route_ratelimit_info(env, _1.defaults[:rateattack])
      end

    [200, {}, StringIO.new({ data: ra_ratelimit_infos }.to_json)]
  end

  def handle_rateattack_request(env, route)
    limit_key = rateattack_request_key(env, route.defaults[:rateattack])
    current_info = get_route_ratelimit_info(env, route.defaults[:rateattack])

    if (current_info[:limit] - current_info[:count]) <= 0
      return [
        429,
        set_rate_limit_and_merge_headers(current_info),
        StringIO.new(I18n.t('general.ratelimit.exceeded'))
      ]
    end

    @status, @headers, @response = @app.call(env)

    if @status >= 200 && @status <= 400
      current_info[:count] =
        Rails.cache.increment(
          limit_key,
          expires_in: route.defaults[:rateattack][:duration],
        )
    end

    [
      @status,
      set_rate_limit_and_merge_headers(current_info, headers: @headers),
      @response,
    ]
  end

  def wrapped_routes
    Rails.application.routes.routes.filter { _1.defaults[:rateattack] }.compact
      .map { ActionDispatch::Routing::RouteWrapper.new(_1) }
  end

  def remove_prefix_postfix_slash(str)
    str.gsub(%r{^\/}, '').gsub(%r{\/$}, '')
  end

  def set_rate_limit_and_merge_headers(rateattack_info, headers: {})
    headers.merge(
      {
        I18n.t('general.ratelimit.headers.remaining') => [
          rateattack_info[:limit] - rateattack_info[:count],
          0,
        ].max.to_s,
        I18n.t('general.ratelimit.headers.limit') =>
          rateattack_info[:limit].to_s,
      },
    )
  end

  def guess_ip(request)
    ip = request.remote_ip || request.ip
    ip = '127.0.0.1' if ip == '::1'
    begin
      IPAddress.valid?(ip)
    rescue StandardError
      nil
    end
  end
end
